#pragma once

#include "bucketQueue.hpp"
#include "openType.hpp"

#include <folly/FBString.h>
#include <folly/Function.h>
#include <folly/Synchronized.h>
#include <folly/executors/CPUThreadPoolExecutor.h>
#include <folly/executors/NamedThreadFactory.h>
#include <folly/io/IOBufQueue.h>
#include <helpers/storageHelper.h>

#include <atomic>
#include <chrono>
#include <queue>
#include <unordered_map>

namespace rtransfer {

class Storage {
public:
    Storage(folly::fbstring storageId,
        one::helpers::StorageHelperPtr storageHelper);

    folly::Future<folly::IOBufQueue> read(one::helpers::FileHandlePtr,
        std::uint64_t reqId, const std::uint64_t offset, const std::size_t size,
        const std::uint8_t priority);

    folly::Future<folly::Unit> write(one::helpers::FileHandlePtr,
        std::uint64_t reqId, const std::uint64_t offset, folly::IOBufQueue buf,
        std::uint8_t priority);

    folly::Future<one::helpers::FileHandlePtr> open(
        folly::fbstring fileId, OpenType openType);

private:
    struct ReadRequest {
        std::uint64_t reqId;
        folly::Promise<folly::IOBufQueue> promise;
        one::helpers::FileHandlePtr handle;
        std::uint64_t offset;
        std::size_t size;

        std::size_t getSize() const { return size; }
    };

    struct WriteRequest {
        std::uint64_t reqId;
        folly::Promise<folly::Unit> promise;
        one::helpers::FileHandlePtr handle;
        std::uint64_t offset;
        folly::IOBufQueue buf;

        std::size_t getSize() const { return buf.chainLength(); }
    };

    template <typename T>
    class ChainQueue {
        using Offset = std::size_t;
        using Handle = one::helpers::FileHandle *;
        using Key = std::pair<Offset, Handle>;

    public:
        using value_type = T;

        folly::fbvector<T> pop()
        {
            DCHECK(!reqs_.empty());

            auto it = reqs_.end();
            while (it == reqs_.end()) {
                auto k = order_.front();
                order_.pop();
                it = reqs_.find(k);
            }

            folly::fbvector<T> res;
            while (it != reqs_.end()) {
                res.emplace_back(std::move(it->second));
                reqs_.erase(it);
                it = reqs_.find(makeNextKey(res.back()));
            }

            return res;
        }

        void push(T val)
        {
            auto key = makeKey(val);
            reqs_.emplace(key, std::move(val));
            order_.emplace(std::move(key));
        }

        bool empty() { return reqs_.empty(); }

    private:
        Key makeKey(const T &val) { return {val.offset, val.handle.get()}; }
        Key makeNextKey(const T &val)
        {
            return {val.offset + val.getSize(), val.handle.get()};
        }

        std::unordered_multimap<Key, T> reqs_;
        std::queue<Key> order_;
    };

    template <typename Ret, typename T>
    folly::Future<Ret> withFile(
        folly::fbstring fileId, OpenType openType, T &&cb);

    void handleRead(const std::size_t bucket);
    void handleWrite(const std::size_t bucket);
    std::string m(const char *);

    template <typename Queue, typename Cb>
    void usingBucket(Queue &queue, const std::size_t bucket, Cb &&cb);

    folly::fbstring storageIdHex_;
    one::helpers::StorageHelperPtr storageHelper_;

    folly::CPUThreadPoolExecutor executor_{
        1, std::make_shared<folly::NamedThreadFactory>("storage")};

    constexpr static const std::size_t maxInProgress_{100};
    std::size_t currentReadsInProgress_{0};
    std::size_t currentWritesInProgress_{0};

    template <typename T>
    struct ReqHash {
        std::size_t operator()(const T &req) const
        {
            return std::hash<void *>{}(static_cast<void *>(req.handle.get()));
        }
    };

    Buckets<ChainQueue<ReadRequest>, ReqHash<ReadRequest>> readQueue_;
    Buckets<ChainQueue<WriteRequest>, ReqHash<WriteRequest>> writeQueue_;
};

}  // namespace rtransfer
