%%%-------------------------------------------------------------------
%%% @author Konrad Zemek
%%% @copyright (C) 2017 ACK CYFRONET AGH
%%% This software is released under the MIT license
%%% cited in 'LICENSE.txt'.
%%% @end
%%%-------------------------------------------------------------------
%%% @doc
%%% @end
%%%-------------------------------------------------------------------
-module(rtransfer_link_request).
-author("Konrad Zemek").

%%%===================================================================
%%% Type definitions
%%%===================================================================

-type priority() :: low_priority | medium_priority
                  | high_priority | very_high_priority.

-type trimmed_notify_fun() ::
        fun((non_neg_integer(), pos_integer()) -> any()).

-type t() :: #{
         provider_id := binary(),
         file_guid := binary(),
         src_storage_id := binary(),
         src_file_id := binary(),
         dest_storage_id := binary(),
         dest_file_id := binary(),
         space_id := binary(),
         offset := non_neg_integer(),
         size := pos_integer(),
         priority := priority(),

         ref => reference(),
         req_id => non_neg_integer(),
         conn => pid(),
         notify => rtransfer:notify_fun(),
         on_complete => rtransfer:on_complete_fun(),
         transfer_data => binary()
        }.

%%%===================================================================
%%% Exports
%%%===================================================================

-export([fetch/2]).
-export_type([t/0, priority/0]).

%%%===================================================================
%%% API
%%%===================================================================

-spec fetch(ConnectionId :: binary(), Req :: t()) -> any().
fetch(ConnectionId, #{} = Req) ->
    spawn(fun() -> do_fetch(ConnectionId, Req) end).

%%%===================================================================
%%% Helpers
%%%===================================================================

-spec do_fetch(ConnectionId :: binary(), Req :: t()) -> any().
do_fetch(ConnectionId, #{on_complete := OnCompleteFun,
                         notify := NotifyFun0,
                         ref := Ref} = Req) ->
    Self = self(),
    spawn(fun() -> guard(Self, Ref, OnCompleteFun) end),
    NotifyFun = fun(O, S) -> NotifyFun0(Ref, O, S) end,
    try
        Wrote = do_fetch(ConnectionId, NotifyFun, Req),
        OnCompleteFun(Ref, {ok, Wrote})
    catch
        Reason -> OnCompleteFun(Ref, {error, Reason})
    end.

-spec do_fetch(ConnectionId :: binary(), trimmed_notify_fun(),
               Req :: t()) -> any().
do_fetch(ConnectionId, NotifyFun,
         #{file_guid := FileGuid,
           src_storage_id := SrcStorageId,
           src_file_id := SrcFileId,
           dest_storage_id := DestStorageId,
           dest_file_id := DestFileId,
           offset := Offset,
           size := Size,
           priority := Priority,
           req_id := ReqId,
           transfer_data := TransferData}) ->
    FetchReq = #{fetch => #{connection_id => ConnectionId,
                            src_storage_id => base64:encode(SrcStorageId),
                            src_file_id => base64:encode(SrcFileId),
                            dest_storage_id => base64:encode(DestStorageId),
                            dest_file_id => base64:encode(DestFileId),
                            file_guid => base64:encode(FileGuid),
                            offset => Offset,
                            size => Size,
                            priority => priority_atom_to_value(Priority),
                            req_id => ReqId,
                            transfer_data => base64:encode(TransferData)}},

    {ok, Handle} = rtransfer_link_callback:open(FileGuid, write),
    schedule_cleanup(
      fun() ->
              rtransfer_link_callback:fsync(Handle),
              rtransfer_link_callback:close(Handle)
      end),
    rtransfer_link_port:request(FetchReq),
    #{<<"wrote">> := Wrote} = handle_fetch_updates(NotifyFun),
    to_integer(Wrote).

-spec handle_fetch_updates(trimmed_notify_fun()) -> term().
handle_fetch_updates(NotifyFun) ->
    Resp = receive {response, _ReqId, R} -> R end,
    ensure(connection, Resp),
    case Resp of
        #{<<"progress">> := Progress} ->
            Offset = to_integer(maps:get(<<"offset">>, Progress, 0)),
            Size = to_integer(maps:get(<<"size">>, Progress, 0)),
            NotifyFun(Offset, Size),
            handle_fetch_updates(NotifyFun);
        Other ->
            Other
    end.

-spec ensure(Type :: atom(), {error, term()} | Other) ->
                    Other | no_return()
                        when Other :: term().
ensure(Type, {error, Reason}) ->
    throw({Type, Reason});
ensure(_Type, Result) ->
    Result.

-spec schedule_cleanup(fun(() -> any())) -> any().
schedule_cleanup(Fun) ->
    Self = self(),
    spawn(fun() -> cleanup(Self, Fun) end).

-spec cleanup(Parent :: pid(), fun(() -> any())) -> any().
cleanup(Parent, Fun) ->
    erlang:monitor(process, Parent),
    receive {'DOWN', _, _, _, _} -> Fun() end.

-spec to_integer(integer() | binary()) -> integer().
to_integer(Val) when is_integer(Val) -> Val;
to_integer(Val) when is_binary(Val) -> binary_to_integer(Val).

-spec guard(Parent :: pid(), Ref :: reference(),
            rtransfer_link:on_complete_fun()) -> any().
guard(Parent, Ref, OnCompleteFun) ->
    erlang:monitor(process, Parent),
    receive
        {'DOWN', _Ref, process, _Pid, normal} -> ok;
        {'DOWN', _Ref, process, _Pid, Reason} ->
            lager:warning("Fetch ~p failed due to ~p~n", [Ref, Reason]),
            OnCompleteFun(Ref, {error, {other, Reason}})
    end.

-spec priority_atom_to_value(atom()) -> 1..4.
priority_atom_to_value(low_priority) -> 4;
priority_atom_to_value(medium_priority) -> 3;
priority_atom_to_value(high_priority) -> 2;
priority_atom_to_value(very_high_priority) -> 1.
